#!/usr/bin/perl -w
# -----------------------------------------------------------------------------
#
# addjailsw.pl
# Add software to the Chrooted environment
# A member of Jail Chroot Project.
# Released under GPL license version 2.0 
# 
# Juan M. Casillas
# assman@gsyc.inf.uc3m.es
#
# This script creates the required files and directories in the chrooted
# environment, and add the software.
#
# -----------------------------------------------------------------------------

use File::Basename;
use File::Copy;
use File::Path;

$LDD = "ldd -v";
$STRACE = "strace";
$LS = "/bin/ls";
$WHEREIS = "/usr/bin/whereis";
$MKNOD = "mknod";

@GLOBAL_DEVICES = ();
@GLOBAL_FILES = ();

$DEBUG = 0;
$INSTALL_PROGRAM = 0;
$INSTALL_PROGRAM_ARGS = 0;

$CREATE_DIR_MASK = 0755;
$COPY_FILE_MASK = 0755;
$TMP_DIR_MASK = 04777;

@NOT_OVERWRITE = ( "/etc/passwd", "/etc/group", "/etc/shadow" );
@STANDARD_TMP_DIRECTORIES = ( "/tmp", "/var/tmp" );

# ----------------------------------------------------------------------------
#
# Programs to be installed in the system
# 
# ----------------------------------------------------------------------------

%PROGRAMS = ();

$PROGRAMS{'bash'} = "";
$PROGRAMS{'sh'} = "";
$PROGRAMS{'cat'} = "";
$PROGRAMS{'grep'} = "";
$PROGRAMS{'id'} = "";
$PROGRAMS{'ls'} = "";
$PROGRAMS{'mkdir'} = "";
$PROGRAMS{'mv'} = "";
$PROGRAMS{'cp'} = "";
$PROGRAMS{'ln'} = "";
$PROGRAMS{'pwd'} = "";
$PROGRAMS{'rm'} = "";
$PROGRAMS{'rmdir'} = "";
$PROGRAMS{'tail'} = "";
$PROGRAMS{'more'} = "";
$PROGRAMS{'touch'} = "";
$PROGRAMS{'vi'} = "-c q";
$PROGRAMS{'head'} = "";

# -----------------------------------------------------------------------------
#
# show_credits()
# show basic credits
#
# -----------------------------------------------------------------------------

sub show_logo {

print "addjailsw.pl\n";
print "A component of Jail\n";
print "http://www.gsyc.inf.uc3m.es/~assman/jail/\n";
print "Juan M. Casillas <assman\@gsyc.inf.uc3m.es>\n\n";

}

# -----------------------------------------------------------------------------
#
# show_usage()
# show basic usage
#
# -----------------------------------------------------------------------------

sub show_usage {
  
  local $prg_name = $_[0];
 
  show_logo();
  print "usage: $prg_name chrootdir [-D] [-P program args] when:\n\n";
  
  print "chrootdir   is the directory to chroot where the software is\n";
  print "            going to be installed\n";
  print "-D          turn the debug on (show the files being copied)\n";
  print "-P          choose to install a given program\n";
  print "   program  program name\n";
  print "   args     command arguments to be passed to the program\n";
  print "            where running the strace (to keep it form get the\n";
  print "            tty an not exit)\n\n";

}

# -----------------------------------------------------------------------------
#
# mkdir_recursive()
# create a directory recursively (works as mkdir -p)
#
# -----------------------------------------------------------------------------

sub mkdir_recursive {

 local $dst_dir = $_[0];

 local $createdir = "/";
 foreach  $i (split(/\//,$dst_dir)) {

   if ($i ne "") {
     $createdir = $createdir.$i;
     mkdir($createdir,$CREATE_DIR_MASK);
     
     $createdir = $createdir."/";
   }
 }

 return(1);
}


# -----------------------------------------------------------------------------
#
# copy_and_create()
# copy a file prepending to the path $2 and build the directory if 
# it doesn't exist It does some checks also to test the validity of
# the path
#
# example:
# copy_and_create("/etc/passwd",/tmp/chroot");
# this will create /tmp/chroot/etc and then copy passwd in this file
#
# -----------------------------------------------------------------------------

sub copy_and_create {

  local $fname = $_[0];
  local $path  = $_[1];
  
  local $base = dirname($fname);  
  local $filen = basename($fname);
  local $dst_dir = "${path}${base}";

  if ( ! -d $dst_dir ) {

    mkdir_recursive($dst_dir) || die("Can't create $dst_dir; bailing out");

    #
    # create all the directories
    # recursively
    # 
       
  }

  #
  # copy the file
  #

  foreach $k (@NOT_OVERWRITE) {
    if ($k eq $fname && -e "$dst_dir/$filen") {
      print("Warning: not allowed to overwrite $dst_dir/$filen \n");
      return;
    }
  }

  if ( -e "$dst_dir/$filen" ) {
    print "Warning: file $dst_dir/$filen exists. Overwritting it\n";
  }

  copy($fname,"$dst_dir/$filen");
  chmod($COPY_FILE_MASK,"$dst_dir/$filen");
}

# -----------------------------------------------------------------------------
#
# lib_dep()
# generate library dependences of a program, so we can 
# insert it and the required libraries into the chrooted
# environment
#
# depedens on ldd and strace
#
# -----------------------------------------------------------------------------

sub lib_dep {
  local $p_name = $_[0];
  local $p_args = $_[1];

  #
  # extract library dependences
  #
  
  local $lib_query = "(".$LDD." ".$p_name." ".$p_args." </dev/null 2>&1)";

  local @output = `$lib_query`;
  foreach $i (@output) {
    if ($i =~ /(.*)=>(.*)/) {
      @r = split(/\s/,$2);
      push(@GLOBAL_FILES,$r[1]);
    }
  }

  #
  # generate file dependencies
  #

  
  local $strace_query= "(".$STRACE." ".$p_name." ".$p_args." </dev/null 2>&1)";

  local @output = `$strace_query`;
  local @filedep = ();
  foreach $i (@output) {
    
    if ($i =~ /open\((.*)\)/ && ($i !~ /= -1 ENOENT/)) {
      @r = split(/,/,$1);
      $r2 = $r[0];
      $r2 =~ s/\"//g;

     push(@filedep,$r2); 
    }

  }

  local @devices = ();
  foreach $i (@filedep) {
    
    local $fname = basename($i);

    if (-c $i) {
      local $dev_query = $LS." -la ".$i;
      local $result = `$dev_query`;

      $result =~ s/( )+/ /g;         # compress spaces
      @fdata = split(/\s/,$result);  #extract fields

      $major = $fdata[4];
      $minor = $fdata[5];
      $major =~ s/,//g;

      local $dev = "$fname:c:".$major.":".$minor;
      push(@GLOBAL_DEVICES,$dev);
    }

    if (-c $i) {
      local $dev_query = $LS." -la ".$i;
      local $result = `$dev_query`;

      $result =~ s/( )+/ /g;         # compress spaces
      @fdata = split(/\s/,$result);  #extract fields

      $major = $fdata[4];
      $minor = $fdata[5];
      $major =~ s/,//g;

      local $dev = "$fname:b:".$major.":".$minor;
      push(@devices,$dev);
    }
    
    if (-f $i) {
      push(@GLOBAL_FILES,$i);
    }
  }
  
return(1);

}


# -----------------------------------------------------------------------------
#
# generate_unique()
# parse the global lists, and build a unique version of them
#
# -----------------------------------------------------------------------------


sub generate_unique {

local %seen = ();
foreach $item (@GLOBAL_FILES) {
    $seen{$item}++;
}
@GLOBAL_FILES = keys %seen;

local %seen = ();
foreach $item (@GLOBAL_DEVICES) {
    $seen{$item}++;
}
@GLOBAL_DEVICES = keys %seen;

}

# -----------------------------------------------------------------------------
#
# locate_file()
# return a path to the file, or false
#
# -----------------------------------------------------------------------------

sub locate_file {
  local $fname = $_[0];
  local $whereis_query = $WHEREIS." ".$fname;
  
  local $result = `$whereis_query`;
  $result =~ s/( )+/ /g;
  @fdata = split(/\s/,$result);

  if ($fdata[1] eq "") {
    return(0);
  }
  
  return($fdata[1]);
}


# -----------------------------------------------------------------------------
#
# print_debug()
# show the database being installed here
#
# -----------------------------------------------------------------------------

sub print_debug {

  print "---[ files ]---\n";

  foreach $i (@GLOBAL_FILES) {
    print "$i\n";
  }

  print "---[ devices ]---\n";
  
  foreach $i (@GLOBAL_DEVICES) {
    print "$i\n";
  }
  
}

# -----------------------------------------------------------------------------
#
# special_devices()
# create the special devices.  syntax is: name:char|block:major:minor
#
# -----------------------------------------------------------------------------

sub special_devices {
  local $item = $_[0];
  local $basedir = $_[1];

  local @things = split(/:/,$item);
  local $dname =  $things[0];
  local $dtype =  $things[1];
  local $dmajor = $things[2];
  local $dminor = $things[3];

  if (! -e "$basedir/dev/$dname") {
    system("$MKNOD $basedir/dev/$dname $dtype $dmajor $dminor");
  }
}

# -----------------------------------------------------------------------------
#
# finish_install()
# this function cleans and change the permissions to the right directories
#
# -----------------------------------------------------------------------------

sub finish_install {
 local $basedir = $_[0];

 for $i (@STANDARD_TMP_DIRECTORIES) {

   if (-e "${basedir}$i") {
     rmtree("${basedir}${i}");
   }
   mkdir_recursive("${basedir}${i}");
   chmod($TMP_DIR_MASK,"${basedir}${i}");   
 }

}



# -----------------------------------------------------------------------------
#
# main loop
#
# -----------------------------------------------------------------------------

if ( $#ARGV + 1 <  1 ) {
  show_usage($0);
  exit(0);
}

#
# parse command switches
#

$chroot_dir = $ARGV[0];

for ($n=1; $n < $#ARGV + 1; $n++) {

  if ($ARGV[$n] eq "-D") {
    $DEBUG = 1;
  }
  
  if ($ARGV[$n] eq "-P") {
    if ($n + 1 < $#ARGV + 1) {
      $INSTALL_PROGRAM = $ARGV[$n+1];
      $n++;
    }
    else {
      show_usage();
      exit(0);
    }
    if ($n + 1 < $#ARGV + 1) {
      $INSTALL_PROGRAM_ARGS = $ARGV[$n+1];
      $n++;
    }
  }
  
}



#
# generate directory & file dependencies
# if we want to install some program, only do that, else
# install the factory default programs
#

show_logo();

if ($INSTALL_PROGRAM) {
    print "Guessing ".$INSTALL_PROGRAM." args(".$INSTALL_PROGRAM_ARGS.")\n";
  $path = locate_file($INSTALL_PROGRAM);
  
  if ($path) {
    push(@GLOBAL_FILES,$path);
    lib_dep($path,$INSTALL_PROGRAM_ARGS);
  }
  else {
    print("Warning: can't locate $INSTALL_PROGRAM");
  }
}
else {
  foreach $i(keys %PROGRAMS) {
    print "Guessing ".$i." args(".$PROGRAMS{$i}.")\n";
    $path = locate_file($i);
    
    if ($path) {
      push(@GLOBAL_FILES,$path);
      lib_dep($path,$PROGRAMS{$i});
    }
    
    else {
      print("Warning: can't locate $i");
    }
  }
}
#
# get only one instance of the libraries
#

generate_unique();

#
# show debug
#

if ($DEBUG) {
  print_debug();
  print("\n\n");
}  

#
# copy the files into the right places
#

foreach $i (@GLOBAL_FILES) {
  print("creating $chroot_dir/$i\n");
  copy_and_create($i,$chroot_dir);
}

foreach $i (@GLOBAL_DEVICES) {
  print("creating $chroot_dir/$i\n");
  special_devices($i,$chroot_dir);
}

finish_install($chroot_dir);

print("\nDone.\n\n");
