#!/usr/bin/perl -w
# -----------------------------------------------------------------------------
#
# mkenv.pl
# generate chroot & user environment
# A member of Jail Chroot Project.
# Released under GPL license version 2.0 
# 
# Juan M. Casillas
# assman@gsyc.inf.uc3m.es
#
# This script generates the chroot environment and the home directory of 
# the chrooted user. This release has been ported to perl to be
# platform indepent, extensible and highly configurable.
#
# -----------------------------------------------------------------------------

use File::Basename;

$MKNOD  = "mknod";
$CREATE_DIR_MASK = 0755;
$SHADOW_FILE_MASK = 0640;


@STANDARD_GENERIC_DIRECTORIES = ( "/dev", "/etc" );


@STANDARD_PASSWD_USERS = ( "root", "daemon", "bin", "sys", "man", "nobody" );

@STANDARD_GROUP_USERS = ( "root", "daemon", "bin", "sys", "adm", "tty",
			  "disk", "kmem", "dialout", "cdrom", "floppy",
			  "tape", "audio", "users", "nogroup" );

#
# added urandom here for proper ssh & scp run
#

@STANDARD_DEVICE_FILES = ( "zero:c:1:5", "null:c:1:3", "urandom:c:1:9" );

# -----------------------------------------------------------------------------
#
# mkdir_recursive()
# create a directory recursively (works as mkdir -p)
#
# -----------------------------------------------------------------------------

sub mkdir_recursive {

 local $dst_dir = $_[0];

 local $createdir = "/";
 foreach  $i (split(/\//,$dst_dir)) {

   if ($i ne "") {
     $createdir = $createdir.$i;
     mkdir($createdir,$CREATE_DIR_MASK);
     
     $createdir = $createdir."/";
   }
 }

 return(1);
}

# -----------------------------------------------------------------------------
#
# show_credits()
# show basic credits
#
# -----------------------------------------------------------------------------

sub show_logo {

print "mkenv.pl\n";
print "A component of Jail\n";
print "http://www.gsyc.inf.uc3m.es/~assman/jail/\n";
print "Juan M. Casillas <assman\@gsyc.inf.uc3m.es>\n\n";

}

# -----------------------------------------------------------------------------
#
# show_usage()
# show basic usage
#
# -----------------------------------------------------------------------------

sub show_usage {
  
  local $prg_name = $_[0];
 
  show_logo();
  print "usage: $prg_name chrootdir when:\n\n";
  print "\tchrootdir\tis the directory to chroot\n\n";

}

# -----------------------------------------------------------------------------
#
# preinstall
# this function generates the STANDARD_DIRECTORIES directories
# under the chrooted environment. 
#
# -----------------------------------------------------------------------------

sub preinstall {
  local $basedir = $_[0];

  foreach $i (@STANDARD_GENERIC_DIRECTORIES) {
    mkdir_recursive("${basedir}${i}");
  }


}

# -----------------------------------------------------------------------------
#
# gen_template_password()
# this function generates the default template passwd, group and shadow
# files under the chrooted environment
#
# -----------------------------------------------------------------------------

sub gen_template_password {
 local $basedir = $_[0];

 local $pass_file = "$basedir/etc/passwd";
 local $grp_file = "$basedir/etc/group";
 local $shw_file = "$basedir/etc/shadow";


# open(GRP,">".$grp_file) || die("Can't open $grp_file: $!");
# open(SHW,">".$shw_file) || die("Can't open $shw_file: $!");

 #
 # Process Passwd File
 #

 open(PASS,"/etc/passwd") || die("Can't open /etc/passwd: $!");
 open(C_PASS,">".$pass_file) || die("Can't open $pass_file: $!");

 while (<PASS>) {
   foreach $i (@STANDARD_PASSWD_USERS) {
     #
     # if matched, send it to the right file
     #
     if (/^$i:/) {
       print C_PASS $_;
     }
   }
 }

 close(PASS);
 close(C_PASS);

 #
 # Process Shadow File
 #

 open(SHW,"/etc/shadow") || die("Can't open /etc/shadow: $!");
 open(C_SHW,">".$shw_file) || die("Can't open $shw_file: $!");

 while (<SHW>) {
   foreach $i (@STANDARD_PASSWD_USERS) {
     #
     # if matched, send it to the right file
     #
     if (/^$i:/) {
       print C_SHW $_;
     }
   }
 }

 close(SHW);
 close(C_SHW);

 #
 # now, change the permissions
 #
 chmod($SHADOW_FILE_MASK, $shw_file);

 #
 # Process group file
 #

 open(GRP,"/etc/group") || die("Can't open /etc/group: $!");
 open(C_GRP,">".$grp_file) || die("Can't open $grp_file: $!");

 while (<GRP>) {
   foreach $i (@STANDARD_GROUP_USERS) {
     #
     # if matched, send it to the right file
     #
     if (/^$i:/) {
       print C_GRP $_;
     }
   }
 }

 close(GRP);
 close(C_GRP);

}

# -----------------------------------------------------------------------------
#
# postinstall()
# some miscellaneous code to execute here. For now, empty
# 
# -----------------------------------------------------------------------------

sub postinstall {
  local $basedir = $_[0];


}

# -----------------------------------------------------------------------------
#
# special_devices()
# create the special devices. Read them from STANDARD_DEVICE_FILES. The syntax
# is: name:char|block:major:minor
#
# -----------------------------------------------------------------------------

sub special_devices {
  local $basedir = $_[0];
  
  foreach $i (@STANDARD_DEVICE_FILES) {

    local @things = split(/:/,$i);
    local $dname =  $things[0];
    local $dtype =  $things[1];
    local $dmajor = $things[2];
    local $dminor = $things[3];

    system("$MKNOD $basedir/dev/$dname $dtype $dmajor $dminor");
  }

}

# -----------------------------------------------------------------------------
#
# main loop
#
# -----------------------------------------------------------------------------


if ( $#ARGV + 1 <  1 ) {
  show_usage($0);
  exit(0);
}

$HDIR = $ARGV[0];

show_logo();
print "Making chrooted environment into $HDIR\n";

print "\tDoing preinstall()\n";
preinstall($HDIR);

print "\tDoing special_devices()\n";
special_devices($HDIR);

print "\tDoing gen_template_password()\n";
gen_template_password($HDIR);

print "\tDoing postinstall()\n";
postinstall($HDIR);


print "Done.\n\n";

exit(1);
